<!---
I'm the centerpoint of the package.  I should be initialized and then
have my ensureCurrent method invoked, which will take care of the rest.

I don't do any work, instead I rely on a collection of migration scripts
(sequentially named migrationNNNNNN.cfc, where N is a digit), each of
which represents a major version of the schema.  Each migration script
must have a 'migrate' method that accepts the schema's current minor
version and does any updates necessary to increase it's minor version as
high as it can go.  Migration scripts will be called in order, starting
with the current major version (in case additional minor versions have
been added), and progressing upwards through all defined migration
scripts.

The semantics of a minor version are entirely up to the individual
migration scripts, but a general rule of thumb is that each one should
be a single DDL statement or a single transaction (of DML).  An abstract
base class (abstractmigration.cfc) has been provided as a starting
point, along with two subclasses: switchbasedmigration.cfc and
sqlbasedmigration.cfc.  Consult their comments for instructions on how
to use them.

------------------------------------------------------------------------
$Id$

Copyright 2007 Barney Boisvert (bboisvert@gmail.com).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
--->
<cfcomponent output="false">

	<cffunction name="init" access="public" output="false" returntype="tool">
		<cfargument name="dsn" type="string" required="true" />
		<cfargument name="migrationPackage" type="string" required="true"
			hint="The package migration classes (named 'migrationNNNNNN.cfc') reside.  Can either be an absolute package name, or a relative (from this package) package name." />
		<cfset variables.my.dsn = dsn />
		<cfset variables.my.migrationPackage = migrationPackage />
		<cfreturn this />
	</cffunction>



	<cffunction name="ensureCurrent" access="public" output="false" returntype="void">
		<cfargument name="lockName" type="string" required="true"
			hint="A unique lock name to use for performing schema updates within.  '##application.applicationname##.schema_tool' is usually a safe bet." />
		<cflock name="#lockName#" type="exclusive" timeout="1" throwontimeout="true">
			<cfset ensureCurrentInternal() />
		</cflock>
	</cffunction>



	<cffunction name="ensureCurrentInternal" access="private" output="false" returntype="void">
		<cfset var state = "" />
		<cfset var state2 = "" />
		<cfset var maxMajor = "" />
		<cfset var major = "" />
		<cfset var migration = "" />
		<cftry>
			<cfset state = getState() />
			<cfcatch type="database">
				<cfif cfcatch.message & cfcatch.detail CONTAINS "schema_version">
					<cfquery datasource="#variables.my.dsn#" name="">
						create table schema_version (
							major_version int not null default 0,
							minor_version int not null default 0
						)
					</cfquery>
					<cfquery datasource="#variables.my.dsn#" name="">
						insert into schema_version
							(major_version, minor_version
							)
						values
							(0,
							 0
							)
					</cfquery>
					<cfset ensureCurrentInternal() />
					<cfreturn />
				<cfelse>
					<cfrethrow />
				</cfif>
			</cfcatch>
		</cftry>
		<!--- at this point, we know what version the schema is at --->
		<cfset maxMajor = getMaxMajorVersion() />
		<cfset minor = state.minorVersion />
		<cfloop from="#max(1, state.majorVersion)#" to="#maxMajor#" index="major">
			<cfset setMajorVersion(major) />
			<cfset migration = getMigrationForVersion(major) />
			<cfset migration.migrate(minor) />
			<cfset minor = 0 />
		</cfloop>
		<cfset state2 = getState() />
		<cfif state.majorVersion NEQ state2.majorVersion
			OR state.minorVersion NEQ state2.minorVersion>
			<cftrace text="Migrated from #state.majorVersion#.#state.minorVersion# to #state2.majorVersion#.#state2.minorVersion#" />
		</cfif>
	</cffunction>
	
	
	
	<cffunction name="getState" access="public" output="false" returntype="struct">
		<cfset var get = "" />
		<cfset var result = structNew() />
		<cfquery datasource="#variables.my.dsn#" name="get">
			select major_version, minor_version
			from schema_version
		</cfquery>
		<cfset result["majorVersion"] = get.major_version />
		<cfset result["minorVersion"] = get.minor_version />
		<cfreturn result />
	</cffunction>
	
	
	
	<cffunction name="getMigrationForVersion" access="private" output="false" returntype="abstractmigration">
		<cfargument name="major" type="numeric" required="true" />
		<cfset var root = "migration" & numberFormat(major, "000000") />
		<cfset var dir = getMigrationPackageDir() />
		<cfif fileExists("#dir#/#root#.sql")>
			<!--- it's SQL based --->
			<cfreturn createObject("component", "sqlbasedmigration").init(variables.my.dsn, "#dir#/#root#.sql") />
		</cfif>
		<cfreturn createObject("component", "#my.migrationPackage#.#root#").init(variables.my.dsn) />
	</cffunction>
	
	
	
	<cffunction name="getMigrationPackageDir" access="private" output="false" returntype="string">
		<cfreturn expandPath("/#replace(my.migrationPackage, '.', '/', 'all')#") />
	</cffunction>
	
	
	
	<cffunction name="getMaxMajorVersion" access="private" output="false" returntype="numeric">
		<cfset var scripts = "" />
		<cfset var dir = getMigrationPackageDir() />
		<cfset var migrations = "" />
		<cfif NOT directoryExists(dir)>
			<!--- not an absolute package, let's see if it's a relative package --->
			<cfset dir = getDirectoryFromPath(getCurrentTemplatePath()) & replace(my.migrationPackage, '.', '/', 'all') />
		</cfif>		
		<cfdirectory action="list"
			directory="#dir#"
			name="migrations"
			sort="name"
			filter="migration??????.???" />
		<cfquery dbtype="query" name="migrations">
			select name
			from migrations
			where name like '%.cfc'
				or name like '%.sql'
			order by name
		</cfquery>
		<cfif migrations.recordCount EQ 0>
			<cfreturn 0 />
		<cfelse>
			<cfreturn val(REReplace(migrations.name[migrations.recordCount], "[^0-9]+", "", "all")) />
		</cfif>
	</cffunction>
	
	
	
	<cffunction name="setMajorVersion" access="private" output="false" returntype="void">
		<cfargument name="majorVersion" type="numeric" required="true" />
		<cfquery datasource="#variables.my.dsn#" name="">
			update schema_version set
				major_version = <cfqueryparam cfsqltype="cf_sql_integer" value="#majorVersion#" />
		</cfquery>
	</cffunction>

<cfoutput></cfoutput>
</cfcomponent>
